use std::io::{stdin, Read, Stdin};

const BS: u8 = 8;
const ESC: u8 = 27;
const TAB: u8 = b'\t';
const LF: u8 = b'\n';
const CR: u8 = b'\r';
const DEL: u8 = 127;

#[derive(Debug)]
pub enum Event {
	Key(Key, Mod),
	Unsupported(Vec<u8>),
	// may add mouse events at some point
}

#[derive(Debug)]
pub enum Mod {
	None,
	Shift,
	Ctrl,
	Alt,
	/// only works for some specific keys
	ShiftCtrl,
	/// only works for some specific keys
	ShiftAlt,
	/// only works for some specific keys
	CtrlAlt,
	/// only works for some specific keys
	ShiftCtrlAlt,
}

#[non_exhaustive]
#[derive(Debug)]
pub enum Key {
	Char(char),
	Right,
	Left,
	Up,
	Down,
	Home,
	End,
	PageUp,
	PageDown,
	/// also Ctrl+H
	Backspace,
	Delete,
	/// also Ctrl+[
	Escape,
	/// \r or \n or Ctrl+M or Ctrl+J
	Enter,
	/// \t or Ctrl+I
	Tab,
	F(u8),
}

pub struct Events {
	source: Stdin,
	buffer: Vec<u8>,
}

impl Events {
	pub fn new() -> Self {
		Self {
			source: stdin(),
			buffer: Vec::new(),
		}
	}

	fn first_byte(&mut self) -> u8 {
		let byte = self.buffer[0];
		self.buffer = self.buffer[1..].to_vec();
		byte
	}

	fn parse(&mut self) -> Event {
		let byte = self.first_byte();
		match byte {
			ESC => self.parse_esc(),
			BS => Event::Key(Key::Backspace, Mod::None),
			DEL => Event::Key(Key::Delete, Mod::None),
			LF | CR => Event::Key(Key::Enter, Mod::None),
			TAB => Event::Key(Key::Tab, Mod::None),
			x @ 1..=26 => Event::Key(Key::Char((x + 96).into()), Mod::Ctrl), // ctrl + letter
			x @ b' '..=b'~' => Event::Key(Key::Char(x.into()), Mod::None),   // regular ascii charaters
			x => Event::Unsupported(vec![x]),
		}
	}

	fn parse_esc(&mut self) -> Event {
		if self.buffer.is_empty() {
			return Event::Key(Key::Escape, Mod::None);
		}

		let byte = self.first_byte();
		match byte {
			ESC => {
				// Two ESC chars in a row probably means that the first one was an escape key press
				// and the second one was for something else, possibly because input is coming all at once from a file.
				// Therefore reinsert the second ESC byte so it can be processed next iteration.
				self.buffer.insert(0, ESC);
				Event::Key(Key::Escape, Mod::None)
			}
			BS => Event::Key(Key::Backspace, Mod::Alt),
			DEL => Event::Key(Key::Delete, Mod::Alt),
			LF | CR => Event::Key(Key::Enter, Mod::Alt),
			TAB => Event::Key(Key::Tab, Mod::Alt),
			x @ 1..=26 => Event::Key(Key::Char((x + 96).into()), Mod::CtrlAlt), // ctrl + letter
			x @ b' '..=b'~' => Event::Key(Key::Char(x.into()), Mod::Alt),       // regular ascii charaters
			x => Event::Unsupported(vec![ESC, x]),
		}
	}
}

impl Iterator for Events {
	type Item = Event;

	fn next(&mut self) -> Option<Self::Item> {
		if self.buffer.is_empty() {
			let mut buf = [0];
			while let Ok(0) = self.source.read(&mut buf) {
				// there has to be a better way to make this blocking...
				std::thread::sleep(std::time::Duration::from_millis(5));
			}
			// TODO handle read errors
			self.buffer.push(buf[0]);
			while let Ok(1) = self.source.read(&mut buf) {
				self.buffer.push(buf[0]);
			}
		}

		Some(self.parse())
	}
}
