#![allow(unused)]
use std::{
	ffi::{c_char, c_int, c_uint},
	io,
	os::fd::{self, AsFd},
};

static mut PREV_TERMINAL_STATE: Option<Termios> = None;

pub fn enter() -> io::Result<()> {
	if unsafe { PREV_TERMINAL_STATE.is_some() } {
		return Ok(());
	}

	let mut state = Termios::default();
	let stdin = io::stdin();

	unsafe {
		// save current terminal state
		check_err(tcgetattr(stdin.as_fd(), &mut state))?;
		PREV_TERMINAL_STATE = Some(state.clone());
	}
	// https://man7.org/linux/man-pages/man3/termios.3.html
	state.c_iflag &= !(IGNBRK | BRKINT | PARMRK | ISTRIP | INLCR | IGNCR | ICRNL | IXON);
	state.c_oflag &= !OPOST;
	state.c_lflag &= !(ECHO | ECHONL | ICANON | ISIG | IEXTEN);
	state.c_cflag &= !(CSIZE | PARENB);
	state.c_cflag |= CS8;
	state.c_cc[VMIN] = 0; // minimum bytes read before read() returns
	state.c_cc[VTIME] = 0; // timeout in 100ms increments (set to none)
	unsafe {
		// set to raw mode
		// cfmakeraw(&state);
		check_err(tcsetattr(stdin.as_fd(), 0, &state))?;
	}
	Ok(())
}

pub fn exit() -> io::Result<()> {
	unsafe {
		if let Some(termios) = PREV_TERMINAL_STATE.as_ref() {
			let stdin = io::stdin();
			check_err(tcsetattr(stdin.as_fd(), 0, termios))?;
		}
	}
	Ok(())
}

fn check_err(val: c_int) -> io::Result<()> {
	if val == -1 {
		Err(io::Error::last_os_error())
	} else {
		Ok(())
	}
}

#[derive(Clone, Default)]
#[repr(C)]
struct Termios {
	pub c_iflag: c_uint,
	pub c_oflag: c_uint,
	pub c_cflag: c_uint,
	pub c_lflag: c_uint,
	pub c_line: c_char,
	pub c_cc: [c_char; 32],
	// these two seem to be platform specific and don't need to be used
	pub c_ispeed: c_int,
	pub c_ospeed: c_int,
}

#[link(name = "c")]
extern "C" {
	/// get terminal flags for fd
	fn tcgetattr(fd: fd::BorrowedFd, termios_p: *mut Termios) -> c_int;
	/// set terminal flags for fd
	fn tcsetattr(fd: fd::BorrowedFd, optional_actions: c_int, termios_p: *const Termios) -> c_int;
	/// set terminal flags to raw mode
	fn cfmakeraw(termios_p: *const Termios);
}
//*
const TCIOFF: c_int = 2;
const TCION: c_int = 3;
const TCOOFF: c_int = 0;
const TCOON: c_int = 1;
const TCIFLUSH: c_int = 0;
const TCOFLUSH: c_int = 1;
const TCIOFLUSH: c_int = 2;
const NL0: c_uint = 0x00000000;
const NL1: c_uint = 0x00000100;
const TAB0: c_uint = 0x00000000;
const CR0: c_uint = 0x00000000;
const FF0: c_uint = 0x00000000;
const BS0: c_uint = 0x00000000;
const VT0: c_uint = 0x00000000;
const VERASE: usize = 2;
const VKILL: usize = 3;
const VINTR: usize = 0;
const VQUIT: usize = 1;
const VLNEXT: usize = 15;
const IGNBRK: c_uint = 0x00000001;
const BRKINT: c_uint = 0x00000002;
const IGNPAR: c_uint = 0x00000004;
const PARMRK: c_uint = 0x00000008;
const INPCK: c_uint = 0x00000010;
const ISTRIP: c_uint = 0x00000020;
const INLCR: c_uint = 0x00000040;
const IGNCR: c_uint = 0x00000080;
const ICRNL: c_uint = 0x00000100;
const IXANY: c_uint = 0x00000800;
const IMAXBEL: c_uint = 0x00002000;
const OPOST: c_uint = 0x1;
const CS5: c_uint = 0x00000000;
const CRTSCTS: c_uint = 0x80000000;
const ECHO: c_uint = 0x00000008;
const OCRNL: c_uint = 0o000010;
const ONOCR: c_uint = 0o000020;
const ONLRET: c_uint = 0o000040;
const OFILL: c_uint = 0o000100;
const OFDEL: c_uint = 0o000200;

const SIGSTKSZ: usize = 8192;
const MINSIGSTKSZ: usize = 2048;
const CBAUD: c_uint = 0o0010017;
const TAB1: c_int = 0x00000800;
const TAB2: c_int = 0x00001000;
const TAB3: c_int = 0x00001800;
const CR1: c_int = 0x00000200;
const CR2: c_int = 0x00000400;
const CR3: c_int = 0x00000600;
const FF1: c_int = 0x00008000;
const BS1: c_int = 0x00002000;
const VT1: c_int = 0x00004000;
const VWERASE: usize = 14;
const VREPRINT: usize = 12;
const VSUSP: usize = 10;
const VSTART: usize = 8;
const VSTOP: usize = 9;
const VDISCARD: usize = 13;
const VTIME: usize = 5;
const IXON: c_uint = 0x00000400;
const IXOFF: c_uint = 0x00001000;
const ONLCR: c_uint = 0x4;
const CSIZE: c_uint = 0x00000030;
const CS6: c_uint = 0x00000010;
const CS7: c_uint = 0x00000020;
const CS8: c_uint = 0x00000030;
const CSTOPB: c_uint = 0x00000040;
const CREAD: c_uint = 0x00000080;
const PARENB: c_uint = 0x00000100;
const PARODD: c_uint = 0x00000200;
const HUPCL: c_uint = 0x00000400;
const CLOCAL: c_uint = 0x00000800;
const ECHOKE: c_uint = 0x00000800;
const ECHOE: c_uint = 0x00000010;
const ECHOK: c_uint = 0x00000020;
const ECHONL: c_uint = 0x00000040;
const ECHOPRT: c_uint = 0x00000400;
const ECHOCTL: c_uint = 0x00000200;
const ISIG: c_uint = 0x00000001;
const ICANON: c_uint = 0x00000002;
const PENDIN: c_uint = 0x00004000;
const NOFLSH: c_uint = 0x00000080;
const CIBAUD: c_uint = 0o02003600000;
const CBAUDEX: c_uint = 0o010000;
const VSWTC: usize = 7;
const OLCUC: c_uint = 0o000002;
const NLDLY: c_uint = 0o000400;
const CRDLY: c_uint = 0o003000;
const TABDLY: c_uint = 0o014000;
const BSDLY: c_uint = 0o020000;
const FFDLY: c_uint = 0o100000;
const VTDLY: c_uint = 0o040000;
const XTABS: c_uint = 0o014000;

const VEOL: usize = 11;
const VEOL2: usize = 16;
const VMIN: usize = 6;
const IEXTEN: c_uint = 0x00008000;
const TOSTOP: c_uint = 0x00000100;
const FLUSHO: c_uint = 0x00001000;
const EXTPROC: c_uint = 0x00010000;
// */
